@@ -68,7 +68,7 @@ import time
 import asyncio
 
 try:
-    from aiohttp import ClientSession, web
+    from aiohttp import ClientSession, ClientTimeout, web
 except ImportError as error:
     print(
         """ImportError: {}:
@@ -206,7 +206,8 @@ class EventListener(EventListenerBase):  # pylint: disable=too-many-instance-att
             if not port:
                 return
             self.address = (ip_address, port)
-            self.session = ClientSession(raise_for_status=True)
+            client_timeout = ClientTimeout(total=10)
+            self.session = ClientSession(raise_for_status=True, timeout=client_timeout)
             self.is_running = True
             log.debug("Event Listener started")
 
